/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.api;

import org.jboss.cache.Cache;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.config.Configuration.NodeLockingScheme;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNull;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.transaction.TransactionManager;
import org.jboss.cache.UnitTestCacheFactory;
import org.jboss.cache.util.TestingUtil;

@Test(groups = {"functional", "jgroups", "pessimistic"}, testName = "api.NodeReplicatedMoveTest")
public class NodeReplicatedMoveTest
{
   private class NodeReplicatedMoveTestTL {
      protected CacheSPI<Object, Object> cache1;
      protected CacheSPI<Object, Object> cache2;
      protected TransactionManager tm;
   }
   protected ThreadLocal<NodeReplicatedMoveTestTL> threadLocal = new ThreadLocal<NodeReplicatedMoveTestTL>();
   
   protected static final Fqn A = Fqn.fromString("/a"), B = Fqn.fromString("/b"), C = Fqn.fromString("/c"), D = Fqn.fromString("/d"), E = Fqn.fromString("/e");
   protected static final Object k = "key", vA = "valueA", vB = "valueB", vC = "valueC", vD = "valueD", vE = "valueE";

   protected NodeLockingScheme nodeLockingScheme = NodeLockingScheme.PESSIMISTIC;

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      NodeReplicatedMoveTestTL tl = new NodeReplicatedMoveTestTL();
      threadLocal.set(tl);
      // start a single cache instance
      tl.cache1 = (CacheSPI<Object, Object>) new UnitTestCacheFactory<Object, Object>().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC), false);
      tl.cache1.getConfiguration().setSyncCommitPhase(true);
      tl.cache1.getConfiguration().setSyncRollbackPhase(true);
      tl.cache1.getConfiguration().setNodeLockingScheme(nodeLockingScheme);
      configure(tl.cache1.getConfiguration());
      tl.cache1.start();
      tl.tm = tl.cache1.getTransactionManager();

      //  start second instance
      tl.cache2 = (CacheSPI<Object, Object>) new UnitTestCacheFactory<Object, Object>().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC), false);
      tl.cache2.getConfiguration().setSyncCommitPhase(true);
      tl.cache2.getConfiguration().setSyncRollbackPhase(true);
      tl.cache2.getConfiguration().setNodeLockingScheme(nodeLockingScheme);
      configure(tl.cache2.getConfiguration());
      tl.cache2.start();
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown()
   {
      NodeReplicatedMoveTestTL tl = threadLocal.get();
      TestingUtil.killCaches(tl.cache1, tl.cache2);      
      threadLocal.set(null);
   }

   protected void configure(Configuration c)
   {
      // to be overridden
   }

   public void testReplicatability()
   {
      NodeReplicatedMoveTestTL tl = threadLocal.get();
      CacheSPI<Object, Object> cache1 = tl.cache1;
      CacheSPI<Object, Object> cache2 = tl.cache2;
      Node<Object, Object> rootNode = tl.cache1.getRoot();
      
      Node<Object, Object> nodeA = rootNode.addChild(A);
      Node<Object, Object> nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(A).getChild(B).get(k));

      assertEquals(vA, cache2.getRoot().getChild(A).get(k));
      assertEquals(vB, cache2.getRoot().getChild(A).getChild(B).get(k));

      // now move...
      cache1.move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(B).get(k));

      assertEquals(vA, cache2.getRoot().getChild(A).get(k));
      assertEquals(vB, cache2.getRoot().getChild(B).get(k));
   }

   public void testInvalidations() throws Exception
   {
      NodeReplicatedMoveTestTL tl = threadLocal.get();
      CacheSPI<Object, Object> cache1 = tl.cache1;
      CacheSPI<Object, Object> cache2 = tl.cache2;
      Node<Object, Object> rootNode = tl.cache1.getRoot();
      cache1.stop();
      cache2.stop();
      cache1.destroy();
      cache2.destroy();
      if (isOptimistic())
      {
         cache1.getConfiguration().setNodeLockingScheme(Configuration.NodeLockingScheme.OPTIMISTIC);
         cache2.getConfiguration().setNodeLockingScheme(Configuration.NodeLockingScheme.OPTIMISTIC);
      }
      cache1.getConfiguration().setCacheMode(Configuration.CacheMode.INVALIDATION_SYNC);
      cache2.getConfiguration().setCacheMode(Configuration.CacheMode.INVALIDATION_SYNC);
      cache1.start();
      cache2.start();

      Node<Object, Object> nodeA = rootNode.addChild(A);
      Node<Object, Object> nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(A).getChild(B).get(k));

      assertInvalidated(cache2, A, "Should be invalidated");
      assertInvalidated(cache2, Fqn.fromRelativeElements(A, B.getLastElement()), "Should be invalidated");

      // now move...
      cache1.move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(B).get(k));

      assertInvalidated(cache2, A, "Should be invalidated");
      assertInvalidated(cache2, B, "Should be invalidated");

      // now make sure a node exists on cache 2
      cache2.getRoot().addChild(A).put("k2", "v2");

      // te invalidation will happen in afterCompletion, hence no exception!
      try
      {
         cache1.move(B, A);// should throw an NPE
         if (!isOptimistic()) assert false : "Should throw an exception!";
      }
      catch (Exception expected)
      {
         if (isOptimistic()) assert false : "Should not have thrown an exception!";
      }
   }

   private void assertInvalidated(Cache cache, Fqn fqn, String msg)
   {
      assert cache.getRoot().getChild(fqn) == null : msg;
      NodeSPI n = ((CacheSPI) cache).peek(fqn, true, true);
      assert n == null || isOptimistic() : msg;
      assert !isOptimistic() || !n.isValid() : msg;
   }

   public void testReplTxCommit() throws Exception
   {
      NodeReplicatedMoveTestTL tl = threadLocal.get();
      CacheSPI<Object, Object> cache1 = tl.cache1;
      CacheSPI<Object, Object> cache2 = tl.cache2;
      Node<Object, Object> rootNode = tl.cache1.getRoot();
      Fqn A_B = Fqn.fromRelativeFqn(A, B);
      Node<Object, Object> nodeA = rootNode.addChild(A);
      Node<Object, Object> nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(A).getChild(B).get(k));

      assertEquals(vA, cache2.getRoot().getChild(A).get(k));
      assertEquals(vB, cache2.getRoot().getChild(A).getChild(B).get(k));

      // now move...
      tl.tm.begin();
      cache1.move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache1.get(A, k));
      assertNull(cache1.get(A_B, k));
      assertEquals(vB, cache1.get(B, k));
      tl.tm.commit();

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(B).get(k));
      assertEquals(vA, cache2.getRoot().getChild(A).get(k));
      assertEquals(vB, cache2.getRoot().getChild(B).get(k));

   }

   public void testReplTxRollback() throws Exception
   {
      NodeReplicatedMoveTestTL tl = threadLocal.get();
      CacheSPI<Object, Object> cache1 = tl.cache1;
      CacheSPI<Object, Object> cache2 = tl.cache2;
      Node<Object, Object> rootNode = tl.cache1.getRoot();
      Node<Object, Object> nodeA = rootNode.addChild(A);
      Node<Object, Object> nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(A).getChild(B).get(k));
      assertEquals(vA, cache2.getRoot().getChild(A).get(k));
      assertEquals(vB, cache2.getRoot().getChild(A).getChild(B).get(k));

      // now move...
      tl.tm.begin();
      cache1.move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache1.get(A, k));
      assertEquals(vB, cache1.get(B, k));

      tl.tm.rollback();

      assertEquals(vA, cache1.getRoot().getChild(A).get(k));
      assertEquals(vB, cache1.getRoot().getChild(A).getChild(B).get(k));
      assertEquals(vA, cache2.getRoot().getChild(A).get(k));
      assertEquals(vB, cache2.getRoot().getChild(A).getChild(B).get(k));
   }

   protected boolean isOptimistic()
   {
      return nodeLockingScheme == NodeLockingScheme.OPTIMISTIC;
   }
}
